#include <V_3D.h>




V_3D::V_3D()
{
  for(int k_dim=0;k_dim<3;k_dim++)
    V[k_dim]=0;
}

V_3D::~V_3D()
{}

V_3D::V_3D(const V_3D &v_3d)
{
  for(int k_dim=0;k_dim<3;k_dim++)
    V[k_dim] = v_3d.V[k_dim];
}

V_3D::V_3D(double x,double y,double z)
{
  V[0]=x;
  V[1]=y;
  V[2]=z;
}

int V_3D::set(double *_v)
{
  for(int k_dim=0;k_dim<3;k_dim++)
    V[k_dim]=_v[k_dim];
  return 0;
}
int V_3D::set(int k_dim,double value)
{
  V[k_dim]=value;
  return 0;
}

int V_3D::set(V_3D _V)
{
  for(int k_dim=0;k_dim<3;k_dim++)
    V[k_dim]=_V.get(k_dim);
  return 0;
}

int V_3D::set(double x,double y,double z)
{
  V[0] = x;
  V[1] = y;
  V[2] = z;
  return 0;
}
  
double V_3D::get(int k_dim) const
{
  return V[k_dim];
}

int V_3D::get(double *_v) const
{
  for(int k_dim=0;k_dim<3;k_dim++)
    _v[k_dim]=V[k_dim];
  return 0;
}

V_3D V_3D::get() const
{
  return *this;
}

double *V_3D::get_pointer()
{
  return V;
}

V_3D operator+(const V_3D& v1,const V_3D& v2)
{
  V_3D Z;
  int k_dim=0;
  for(k_dim=0;k_dim<3;k_dim++)
    Z.set(k_dim,v1.V[k_dim]+v2.V[k_dim]);
  return Z;
}

V_3D operator-(const V_3D& v1,const V_3D& v2)
{
  V_3D Z;
  int k_dim=0;
  for(k_dim=0;k_dim<3;k_dim++)
    Z.set(k_dim,v1.V[k_dim]-v2.V[k_dim]);
  return Z;
}

V_3D operator*(const double& alpha,const V_3D& v1)
{
  V_3D Z;
  int k_dim=0;
  for(k_dim=0;k_dim<3;k_dim++)
    Z.set(k_dim,alpha*v1.V[k_dim]);
  return Z;
}

V_3D operator*(const V_3D& v1,const double& alpha)
{
  V_3D Z;
  int k_dim=0;
  for(k_dim=0;k_dim<3;k_dim++)
    Z.set(k_dim,alpha*v1.V[k_dim]);
  return Z;
}


V_3D operator/(const V_3D& v1,const double& alpha)
{
  V_3D Z;
  int k_dim=0;
  for(k_dim=0;k_dim<3;k_dim++)
    Z.set(k_dim,v1.V[k_dim]/alpha);
  return Z;
}


double V_3D::dot(V_3D _v) const
{
  double res=0.0;
  int k_dim=0;
  for(k_dim=0;k_dim<3;k_dim++)
    res += V[k_dim]*_v.get(k_dim);
  return res;
}

double V_3D::norm() const
{
  double res=0.0;
  int k_dim=0;
  for(k_dim=0;k_dim<3;k_dim++)
    res += V[k_dim]*V[k_dim];
  
  return powf(res,0.5);
}

V_3D& V_3D::operator=(const V_3D& _v)
{
  for(int k_dim=0;k_dim<3;k_dim++)
    V[k_dim] = _v.V[k_dim];
  return *this;
}

V_3D V_3D::vector_prod(V_3D _v) const
{
  V_3D res;

  res.set(0,V[1]*_v.get(2)-V[2]*_v.get(1));
  res.set(1,V[2]*_v.get(0)-V[0]*_v.get(2));
  res.set(2,V[0]*_v.get(1)-V[1]*_v.get(0));
  
  return res;
}

double V_3D::area(V_3D _v) const
{
  return vector_prod(_v).norm()/2;
}


ostream& operator << (ostream& flux, V_3D _v)
{
  flux<<"("<<_v.get(0)<<","<<_v.get(1)<<","<<_v.get(2)<<")";
  return flux;
}

int operator==(const V_3D& v1,const V_3D& v2)
{
  double epsilon=0.0001;
  if(fabs(v1.V[0]-v2.V[0])<epsilon
     && fabs(v1.V[1]-v2.V[1])<epsilon
     && fabs(v1.V[2]-v2.V[2])<epsilon)
    return 1;
  else
    return 0;
}

int operator!=(const V_3D& v1,const V_3D& v2)
{
  double epsilon=0.0001;
  if(fabs(v1.V[0]-v2.V[0])>=epsilon
     || fabs(v1.V[1]-v2.V[1])>=epsilon
     || fabs(v1.V[2]-v2.V[2])>=epsilon)
    return 1;
  else
    return 0;
}



double& V_3D::operator[](int index)
{
  if(index<0 || index>2)
    {printf("error in operator [] in V_3d, index [%d] not correct\n",index);exit(-1);}
  return V[index];
}
double V_3D::operator[](int index) const
{
  if(index<0 || index>2)
    {printf("error in operator [] in V_3d, index [%d] not correct\n",index);exit(-1);}
  return V[index];
}

V_3D& V_3D::operator+=(const V_3D& _v)
{
  V[0] += _v[0];
  V[1] += _v[1];
  V[2] += _v[2];
  return *this;
}
V_3D& V_3D::operator/=(const double& alpha)
{
  for(int k_dim=0;k_dim<3;k_dim++)
    V[k_dim] /= alpha;
  return *this;
}

V_3D V_3D::get_min_scalar(const V_3D& _v) const
{
  V_3D temp;
  for(int k_dim=0;k_dim<3;k_dim++)
    temp[k_dim]=(V[k_dim]<_v[k_dim]?V[k_dim]:_v[k_dim]);
  return temp;
}
V_3D V_3D::get_max_scalar(const V_3D& _v) const
{
  V_3D temp;
  for(int k_dim=0;k_dim<3;k_dim++)
    temp[k_dim]=(V[k_dim]>_v[k_dim]?V[k_dim]:_v[k_dim]);
  return temp;
}

double V_3D::get_max_of_coeff() const
{return max(max(V[0],V[1]),V[2]);}
double V_3D::get_min_of_coeff() const
{return min(min(V[0],V[1]),V[2]);}


