//skinning



#include <stdlib.h>
#include <iostream>
#include <GL/glut.h>
#include <cmath>
#include <string>
#include <errno.h>
#include <string>
#include <GL/glext.h>
#include <stdio.h>

#include <vector>
using namespace std;
#define PI 3.14159



#include <Mesh.h>
#include <File_parser.h>
#include <OpenGL_drawer.h>
#include <Navigator_tool.h>
#include <Skeleton.h>
#include <Skinned.h>
#include <subdivision.h>
#include <double_vector.h>


//glut functions
static void display_callback();
static void reshape_callback (int width, int height);
static void keyboard_callback(unsigned char key,int x,int y);
static void idle_callback(int value);
static void motion_callback (int x, int y);
static void mouse_callback (int button, int action, int x, int y);


//main functions
int setup_opengl(int argc,char **argv);
int draw_scene(); //draw the model
int load_model(); //load the model
int init_skeleton(); //init size of the skeleton
//calculate multiresolution animation
int load_animation(int k_bone,int k_dim,double A,double B,double C);
//build multiresolution for angle functions
int build_time_multires(const std::vector <double> angle,std::vector <double_vector> *average,std::vector<double_vector> *residual);
double get_rand();//just a rand



//GLOBAL VARIABLES
//************************************//
OpenGL_drawer openGL_drawer;
Navigator_tool navigator;


//number of space subdivision
int N_subdiv=4;

//number of time subdivision
int N_time_subdiv=8; //(automatically adjusted with space threshold)

//number of meshes
int N_mesh=17;

//skinned meshes
std::vector <Skinned> skinned(N_subdiv);

//threshold to show resolutions
std::vector <double> distance_threshold(N_subdiv-1);

//threshold for time animation
std::vector <double> distance_time_threshold(N_time_subdiv-1);

//positions of meshes
std::vector <V_3D> init_position(N_mesh);

//skeleton with there animations
std::vector <Skeleton> skeleton_vector(N_mesh*N_time_subdiv);

//skeleton which are animated
std::vector <Skeleton> deformed_skeleton_vector(N_mesh*N_time_subdiv);













int main(int argc, char *argv[])
{
  cout<<"**hello**\n"<<endl;
  printf("run %s with %d parameters ... \n\n",argv[0],argc-1);


  //start OpenGL
  //*************//
  navigator.set_screen_size(800,800);
  setup_opengl(argc,argv);



  //load  models
  //*************//
  load_model();



  //animation
  //*************//
  init_skeleton();
  //set-up manually the animation
  load_animation(1,2,1.0,0.4,0.1);//(k_bone,k_dim,A,B,C,D)
  load_animation(1,1,1.0,0.0,0.0);
  load_animation(1,0,1.0,0.0,0.0);
  load_animation(4,0,-1.0,0.4,0.1);
  load_animation(4,2,-1.0,0.4,0.1);



  //init positions
  //*************//
  double L=10.0;
  for(int k_mesh=0;k_mesh<N_mesh;k_mesh++)
    init_position[k_mesh].set(get_rand()*L,get_rand(),get_rand()*L);




  //set distance threshold for multiresolution
  //*************//
  distance_threshold[N_subdiv-2]=1.0;
  for(int k=N_subdiv-3;k>=0;k--)
    distance_threshold[k]=distance_threshold[k+1]*2.0;
  //multiply distance by 2 at each steps

  


  //set distance threshold for time parameter
  //*************//
  for(int k=0;k<N_time_subdiv-1;k++)
    {
      //linear cut
      double u=double(k)/double(N_time_subdiv-1.0);
      distance_time_threshold[k]=4*distance_threshold[0]*(1.0-u)+distance_threshold[N_subdiv-2]*u;
    }




  //speed of the animation
  navigator.set_N_speed(250);


  //run
  printf("**RUN**\n\n");
  printf("Navigation with keys:\n\n");
  printf("==========================\n");
  printf("Space Displacement\n");
  printf("  W    R\n");
  printf("A S D  F\n\n");
  printf("and mouse with drag & drop\n\n");
  printf("W: wireframe mode\n");
  printf("F: filled mode\n");
  printf("S: enable/disable skeleton drawing\n");
  printf("==========================\n");
  glutMainLoop ();


  return 0;
}











//Draw the scene
int draw_scene()
{

  glEnable(GL_LIGHTING);

  //curent skeleton and deformed skeleton
  Skeleton *skeleton=NULL;
  Skeleton *deformed_skeleton=NULL;

  
  //current position in space
  V_3D position = -1*navigator.get_current_position();


  //distance of current position to the meshes
  double distance_to_mesh=0.0;


  //LOOP OVER EVERY MESHES
  //****************************************//
  int k_mesh=0;
  for(k_mesh=0;k_mesh<N_mesh;k_mesh++)
    {

      //current distance to the given mesh
      distance_to_mesh = (position-init_position[k_mesh]).norm();


      //Find multiresolution place of this skeleton and mesh//
      //*****************************************//
      
      //find corresponding multiresolution in time
      int k_time_threshold=0;
      while(k_time_threshold<N_time_subdiv-1 && distance_to_mesh<distance_time_threshold[k_time_threshold])
	k_time_threshold++;

      //find corresponding multiresolution in space
      int k_threshold=0;
      while(k_threshold<N_subdiv-1 && distance_to_mesh<distance_threshold[k_threshold])
	k_threshold++;





      //Deform the Skeleton//
      //*****************************************//
      skeleton = &skeleton_vector[k_mesh+N_mesh*(N_time_subdiv-k_time_threshold-1)];
      deformed_skeleton = &deformed_skeleton_vector[k_mesh+N_mesh*(N_time_subdiv-k_time_threshold-1)];
      skeleton->deform_skeleton(navigator.get_time(),deformed_skeleton);
      





      //Deform the Surface//
      //*****************************************//
      skinned[k_threshold].set_deformed_skeleton(deformed_skeleton);//set the corresponding skeleton with the mesh to deform
      skinned[k_threshold].skinning();//skinning




      //Draw Meshes to the correct place//
      //*****************************************//
      glPushMatrix();
      glTranslatef(init_position[k_mesh][0],init_position[k_mesh][1],init_position[k_mesh][2]);//translate
      //draw_it
      openGL_drawer.draw_mesh(*skinned[k_threshold].get_deformed_mesh());
      if(navigator.is_skeleton_shown()==1)
	openGL_drawer.draw_skeleton(*deformed_skeleton);
      glPopMatrix();

    }




  return 0;
}





//load the model
int load_model()
{
  
  File_parser parser;
  char filename[1024]={"data/banana_low.g"};

  int k_subdiv=0;
  Mesh *temp_mesh=NULL;
  Mesh *temp_mesh2=NULL;


  //LOAD MESH and precalculate subdivision    //
  //******************************************//




  for(k_subdiv=0;k_subdiv<N_subdiv;k_subdiv++)
    {
      temp_mesh=new Mesh;
      
      if(k_subdiv==0)//special case (seg fault if subdivide_mesh is called with 0 subdivision)
	parser.load_g_mesh(filename,temp_mesh);
      else
	subdivide_mesh(2,k_subdiv,filename,temp_mesh);

      //load mesh into the skinning class
      skinned[k_subdiv].set_mesh(temp_mesh);

    }


  

  //LOAD SKELETON                             //
  //******************************************//
  Skeleton *temp_skeleton=NULL;
  for(k_subdiv=0;k_subdiv<N_subdiv;k_subdiv++)
    {
      temp_skeleton=new Skeleton;
      parser.load_collada_skeleton("data/skeleton.dae","0",temp_skeleton);


      //load skeleton into the skinning class
      skinned[k_subdiv].set_skeleton(temp_skeleton);
    }






  //SET UP Mesh, skinning weights and Skeleton//
  //******************************************//

  Skeleton *current_skeleton=NULL;
  Skeleton *current_deformed_skeleton=NULL;

  //for every subdivision in space
  for(k_subdiv=0;k_subdiv<N_subdiv;k_subdiv++)
    {
      //init surfaces
      skinned[k_subdiv].get_mesh()->build_smooth_normal();
      skinned[k_subdiv].get_mesh()->fill_color(0.9,0.9,0.5,1.0);

      //init deformed mesh
      temp_mesh2=new Mesh;
      *temp_mesh2=*skinned[k_subdiv].get_mesh();
      skinned[k_subdiv].set_deformed_mesh(temp_mesh2);

      //shorthand
      current_skeleton          = skinned[k_subdiv].get_skeleton();
      current_deformed_skeleton = new Skeleton();
      skinned[k_subdiv].set_deformed_skeleton(current_deformed_skeleton);

      //set matrices
      current_skeleton->fix_init_bind_pose();
      current_skeleton->read_skeleton();
      current_skeleton->precalculate_matrix();

      //skinning weight
      skinned[k_subdiv].automatic_weight_calculation();


      //initialise deformed_skeleton
      *current_deformed_skeleton = *current_skeleton;

    }

  //set time step
  navigator.automatic_set_time(*current_skeleton);




  return 0;
}









//load animation of every skeleton and pre-compute multires
int load_animation(int k_bone,int k_dim,double A,double B,double C)
{

  int k_skeleton=0;
  int k_time_subdiv=0;
  int current_index=0;
  int N_motion=int(powf(2.0,double(N_time_subdiv+1)));

  //vector of angle at better resolution
  std::vector <double> motion_angle(N_motion);

  //multiresolution functions
  std::vector <double_vector> multires_angle(N_motion* N_time_subdiv);
  std::vector <double_vector> multires_angle_residual(N_motion* N_time_subdiv);


  //***************************************//
  //build angle function                   //
  //                                       //
  //CAN BE ANY FUNCTIONS                   //
  //***************************************//
  double u=0.0;int k=0;
  for(k=0,u=0.0;k<N_motion;k++,u=double(k)/double(N_motion-1))
    motion_angle[k] = 20*(A*sin(2*u*PI)+B*cos(8.4*u*PI)+C*sin(12*u*PI));




  //build multiresolution of angle vector
  build_time_multires(motion_angle,&multires_angle,&multires_angle_residual);





  //fill skeletons with animation
  for(k_skeleton=0;k_skeleton<N_mesh;k_skeleton++)
    {
      double rand=2.0*(2.0*get_rand()-0.5);

      for(k_time_subdiv=0;k_time_subdiv<N_time_subdiv;k_time_subdiv++)
	{
	  int size_motion=multires_angle[k_time_subdiv].size();
	  current_index=k_skeleton+N_mesh*k_time_subdiv;

	  //fill the multiresolution angle
	  for(int k_motion=0;k_motion<size_motion;k_motion++)
	    skeleton_vector[current_index].set_angle(k_bone,k_motion,k_dim,rand*multires_angle[k_time_subdiv][k_motion]);
	}
   }

  return 0;
      
}




//set-up Multiresolution in time for every skeleton_vector
int build_time_multires(const std::vector <double> angle,std::vector <double_vector> *average,std::vector<double_vector> *residual)
{
  average->resize(N_time_subdiv);
  residual->resize(N_time_subdiv);
  
  int k_subdiv=0;
  int k_time=0;
  int current_size=angle.size();
  (*average)[0].set(angle);


  //MULTIRESOLUTION                                  //
  //*************************************************//
  for(k_subdiv=1;k_subdiv<N_time_subdiv;k_subdiv++)
    {
      current_size/=2;
      for(k_time=0;k_time<current_size;k_time++)
	{
	  (*average)[k_subdiv].add(
				   0.5*
				   (
				    (*average)[k_subdiv-1][2*k_time+0]
				    +
				    (*average)[k_subdiv-1][2*k_time+1]
				    )
				   );
	  (*residual)[k_subdiv].add(
				    (*average)[k_subdiv-1][2*k_time+1]
				    -
				    (*average)[k_subdiv-1][2*k_time+0]
				    );
	}

    }
  



  //symmetrize for animation
  //*************************************************//
  for(k_subdiv=0;k_subdiv<N_time_subdiv;k_subdiv++)
    {
      current_size=(*average)[k_subdiv].size();
      for(k_time=0;k_time<current_size;k_time++)
	(*average)[k_subdiv].add((*average)[k_subdiv][current_size-k_time-1]);
    }
      
  
  


  return 0;
}











//initialise size of skeleton_vectors with the multiresolution
int init_skeleton()
{
  Skeleton temp;
  int max_size=2*int(powf(2.0,double(N_time_subdiv+1)));//2* => symetric
  int current_size=0;

  int k_skeleton=0,k_time_subdiv=0,current_index=0;

  temp = *(skinned[0].get_skeleton());

  //for every skeleton
  for(k_skeleton=0;k_skeleton<N_mesh;k_skeleton++)
    {
      //this size is divided by two at each multiresolution level
      current_size=max_size;

      //for every time subdivision
      for(k_time_subdiv=0;k_time_subdiv<N_time_subdiv;k_time_subdiv++)
	{
	  current_index=k_skeleton+N_mesh*k_time_subdiv;

	  //init skeleton
	  skeleton_vector[current_index]=temp;
	  deformed_skeleton_vector[current_index]=temp;

	  //set animation size, get ready for multiresolution
	  for(int k=0;k<skeleton_vector[current_index].get_N_joint();k++)
	    {
	      skeleton_vector[current_index].set_size_time(k,current_size);

	      //for every time
	      for(int k_motion=0;k_motion<current_size;k_motion++)
		{
		  //set small random
		  double random_to_add = 0.0;

		  //keep 0.0 and 1.0 without random
		  if(k_motion!=0 && k_motion!=current_size-1)
		    random_to_add=0.5*cos(k_time_subdiv*53453);//not really a random, but ...
		  
		  skeleton_vector[current_index].set_time(k,k_motion,double(k_motion+random_to_add)/double(current_size-1.0));
		}

	    }
	  current_size/=2;//multiresolution size
	}
    }
  navigator.automatic_set_time(skeleton_vector[0]);

  return 0;
}
















//*******************//
//GLUT STUFF
//*******************//








//glut functions
int setup_opengl(int argc,char **argv)
{
  openGL_drawer.init_opengl(argc,argv,navigator);

  //callback
  glutDisplayFunc (display_callback);
  glutReshapeFunc (reshape_callback);
  glutKeyboardFunc(keyboard_callback);
  glutTimerFunc(33,idle_callback,1);
  glutMotionFunc (motion_callback);
  glutMouseFunc (mouse_callback);

  return 0;
}



//display
static void display_callback()
{
  //croping
  double Z_MIN = 0.005;
  double Z_MAX = 300.0;


  //vieport (all screen)
  glViewport (0, 0, navigator.get_X_screen(),navigator.get_Y_screen());
  glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  //Projection
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity ();
  gluPerspective (45, navigator.get_screen_ratio(), Z_MIN,Z_MAX);

  //Modelview
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  glTranslatef(0,0,navigator.get_dist_obs());
  glRotatef(navigator.get_theta_1(),1.0,0.0,0.0);
  glRotatef(navigator.get_theta_2(),0.0,1.0,0.0);
  glTranslatef(navigator.get_tr_x(),navigator.get_tr_y(),navigator.get_tr_z());


  //Draw the scene
  draw_scene();

  //draw the basis
  openGL_drawer.draw_local_basis(navigator);






  //code d'erreur OpenGL
  GLenum errCode;
  const GLubyte *errString;
  if((errCode=glGetError()) != GL_NO_ERROR)
    {
      errString = gluErrorString(errCode);
      printf("ERROR OPENGL : %s\n",errString);
      exit(-2);
    }

  //redraw
  glFlush ();
  glutSwapBuffers ();

}


static void reshape_callback (int width, int height)
{
  navigator.set_screen_size(width,height);
}


//keyboard
static void keyboard_callback(unsigned char key,int x,int y)
{

  x=0;y=0;
  double d_L=0.001*(1+100*abs(navigator.get_dist_obs()));

  #define GLUT_SPACE 32


  switch(key)
    {
    case 'Q'://quit
      printf("\n[EXIT OK]\n\n");
      exit(0);
      break;


    case 'w':
      navigator.go_forward(d_L);
      break;
    case 's':
      navigator.go_backward(d_L);
      break;
    case 'a':
      navigator.go_left(d_L);
      break;
    case 'd':
      navigator.go_right(d_L);
      break;
    case 'r':
      navigator.go_up(d_L);
      break;
    case 'f':
      navigator.go_down(d_L);
      break;

    case 'W'://wireframe
      glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
      break;
    case 'F'://Filled
      glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
      break;
    case 'S'://show skeleton
      navigator.is_skeleton_shown()==1?navigator.set_skeleton_shown(0):navigator.set_skeleton_shown(1);
      break;
    }

  glutPostRedisplay ();
}


static void idle_callback(int value)
{
  value=1;//useless

  navigator.increase_time();
  glutPostRedisplay ();
  glutTimerFunc( 33, idle_callback, 1);
}



static void motion_callback (int x, int y)
{
  /* si le bouton gauche est presse */
  double d_x=x-navigator.get_x_prec(),d_y=y-navigator.get_y_prec();
  if (navigator.get_left_b()) 
    {
      if(navigator.get_current_view()==1)
	{
	  navigator.increase_theta_2(d_x);
	  navigator.increase_theta_1(d_y);
	}
      navigator.set_x_prec(x);
      navigator.set_y_prec(y);
    }

  /* si le bouton droit est presse */
  if (navigator.get_right_b()) 
    {
      double u = double(y-navigator.get_y_prec());
      double fu = abs(u)/500.0;
      if(u<0)
	fu=-fu;
      
      navigator.increase_dist_obs(fu);
      navigator.set_y_prec(y);
    }

  glutPostRedisplay ();
}
static void mouse_callback (int button, int action, int x, int y)
{

  if (button == GLUT_LEFT_BUTTON && action == GLUT_DOWN) {
    navigator.set_left_b(1);
    navigator.set_x_prec(x); 
    navigator.set_y_prec(y);
  }
  if (button == GLUT_LEFT_BUTTON && action == GLUT_UP) 
    navigator.set_left_b(0); 

  if (button == GLUT_RIGHT_BUTTON && action == GLUT_DOWN) {
    navigator.set_right_b(1);
    navigator.set_y_prec(y);
  }
  if (button == GLUT_RIGHT_BUTTON && action == GLUT_UP) 
    navigator.set_right_b(0); 

  glutPostRedisplay ();
}

//just a rand number \in [0,1]
double get_rand()
{
  int precision=50000;
  return double(rand()%precision)/double(precision);
}
